@@ -15,6 +15,7 @@ from dvc.exceptions import (
     DvcIgnoreInCollectedDirError,
     RemoteCacheRequiredError,
 )
+from dvc.hash_info import HashInfo
 from dvc.ignore import DvcIgnore
 from dvc.path_info import PathInfo, URLInfo
 from dvc.progress import Tqdm
@@ -260,17 +261,17 @@ class BaseTree:
             hash_ = None
 
         if hash_:
-            return hash_
+            return HashInfo(self.PARAM_CHECKSUM, hash_)
 
         if self.isdir(path_info):
-            hash_ = self.get_dir_hash(path_info, **kwargs)
+            hash_info = self.get_dir_hash(path_info, **kwargs)
         else:
-            hash_ = self.get_file_hash(path_info)
+            hash_info = self.get_file_hash(path_info)
 
-        if hash_ and self.exists(path_info):
-            self.state.save(path_info, hash_)
+        if hash_info and self.exists(path_info):
+            self.state.save(path_info, hash_info.value)
 
-        return hash_
+        return hash_info
 
     def get_file_hash(self, path_info):
         raise NotImplementedError
@@ -280,7 +281,7 @@ class BaseTree:
             raise RemoteCacheRequiredError(path_info)
 
         dir_info = self._collect_dir(path_info, **kwargs)
-        return self._save_dir_info(dir_info)
+        return self.save_dir_info(dir_info)
 
     def hash_to_path_info(self, hash_):
         return self.path_info / hash_[0:2] / hash_[2:]
@@ -294,7 +295,8 @@ class BaseTree:
         return "".join(parts)
 
     def save_info(self, path_info, **kwargs):
-        return {self.PARAM_CHECKSUM: self.get_hash(path_info, **kwargs)}
+        hash_info = self.get_hash(path_info, **kwargs)
+        return {hash_info.name: hash_info.value}
 
     def _calculate_hashes(self, file_infos):
         file_infos = list(file_infos)
@@ -305,9 +307,8 @@ class BaseTree:
         ) as pbar:
             worker = pbar.wrap_fn(self.get_file_hash)
             with ThreadPoolExecutor(max_workers=self.hash_jobs) as executor:
-                tasks = executor.map(worker, file_infos)
-                hashes = dict(zip(file_infos, tasks))
-        return hashes
+                hashes = (hi.value for hi in executor.map(worker, file_infos))
+                return dict(zip(file_infos, hashes))
 
     def _collect_dir(self, path_info, **kwargs):
         file_infos = set()
@@ -343,20 +344,23 @@ class BaseTree:
         # Sorting the list by path to ensure reproducibility
         return sorted(result, key=itemgetter(self.PARAM_RELPATH))
 
-    def _save_dir_info(self, dir_info):
-        hash_, tmp_info = self._get_dir_info_hash(dir_info)
-        new_info = self.cache.tree.hash_to_path_info(hash_)
-        if self.cache.changed_cache_file(hash_):
+    def save_dir_info(self, dir_info):
+        hash_info, tmp_info = self._get_dir_info_hash(dir_info)
+        new_info = self.cache.tree.hash_to_path_info(hash_info.value)
+        if self.cache.changed_cache_file(hash_info.value):
             self.cache.tree.makedirs(new_info.parent)
             self.cache.tree.move(
                 tmp_info, new_info, mode=self.cache.CACHE_MODE
             )
 
-        self.state.save(new_info, hash_)
+        self.state.save(new_info, hash_info.value)
 
-        return hash_
+        return hash_info
 
     def _get_dir_info_hash(self, dir_info):
+        # Sorting the list by path to ensure reproducibility
+        dir_info = sorted(dir_info, key=itemgetter(self.PARAM_RELPATH))
+
         tmp = tempfile.NamedTemporaryFile(delete=False).name
         with open(tmp, "w+") as fobj:
             json.dump(dir_info, fobj, sort_keys=True)
@@ -366,8 +370,9 @@ class BaseTree:
         to_info = tree.path_info / tmp_fname("")
         tree.upload(from_info, to_info, no_progress_bar=True)
 
-        hash_ = tree.get_file_hash(to_info) + self.CHECKSUM_DIR_SUFFIX
-        return hash_, to_info
+        hash_info = tree.get_file_hash(to_info)
+        hash_info.value += self.CHECKSUM_DIR_SUFFIX
+        return hash_info, to_info
 
     def upload(self, from_info, to_info, name=None, no_progress_bar=False):
         if not hasattr(self, "_upload"):
